package net.laparola.ui;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

public class LaParolaSegnalibri {
	public static class GruppoSegnalibri implements Comparable<GruppoSegnalibri> {
		public String descrizione;
		public String nome;
		public List<LaParolaSegnalibri.Segnalibro> segnalibri;

		public GruppoSegnalibri() {
			segnalibri = new ArrayList<LaParolaSegnalibri.Segnalibro>();
		}

		public GruppoSegnalibri(String nome) {
			this();
			this.nome = nome;
			this.descrizione = "";
		}

		@Override
		public int compareTo(GruppoSegnalibri another) {
			return nome.compareTo(another.nome);
		}
	}

	public static class Segnalibro implements Comparable<Segnalibro> {
		public String nome;
		public List<String> riferimenti;
		public String inizioSottogruppo;

		public Segnalibro() {
			riferimenti = new ArrayList<String>();
		}

		public Segnalibro(String nome, String url) {
			this();
			this.nome = nome;
			this.riferimenti.add(url);
		}

		@Override
		public int compareTo(Segnalibro another) {
			return nome.compareTo(another.nome);
		}
	}

	private class SegnalibriHandler extends DefaultHandler {
		private boolean inDescrizione;
		private boolean inGruppo;
		private boolean inSottoGruppo;
		private boolean inNome;
		private boolean inRiferimento;
		private boolean inSegnalibro;
		private StringBuilder caratteri = new StringBuilder();

		private LaParolaSegnalibri.GruppoSegnalibri ultimoGruppo;
		private LaParolaSegnalibri.Segnalibro ultimoSegnalibro;
		private String ultimoNomeSottoGruppo;

		@Override
		public void characters(char ch[], int start, int length) {
			/*
			 * if (caratteri.length() > 0) { caratteri.append(" "); }
			 */

			caratteri.append(ch, start, length);
		}

		@Override
		public void endDocument() {
			//
		}

		@Override
		public void endElement(String namespaceURI, String localName, String qName) {
			String valore = caratteri.toString().replace('\t', ' ').replace('\r', ' ').replace('\n', ' ').trim();

			if (inNome) {
				if (inSegnalibro) {
					ultimoSegnalibro.nome = valore;
				} else if (inSottoGruppo) {
					ultimoNomeSottoGruppo = valore;
				} else if (inGruppo) {
					ultimoGruppo.nome = valore;
				}
			} else if (inDescrizione) {
				if (inGruppo) {
					ultimoGruppo.descrizione = valore;
				}
			} else if (inRiferimento) {
				if (inSegnalibro) {
					ultimoSegnalibro.riferimenti.add(valore);
				}
			}
			caratteri.setLength(0);

			if (localName.equals("gruppo")) {
				inGruppo = false;
			} else if (localName.equals("sottogruppo")) {
				inSottoGruppo = false;
				ultimoNomeSottoGruppo = null;
			} else if (localName.equals("segnalibro")) {
				inSegnalibro = false;
			} else if (localName.equals("nome")) {
				inNome = false;
			} else if (localName.equals("descrizione")) {
				inDescrizione = false;
			} else if (localName.equals("riferimento")) {
				inRiferimento = false;
			}
		}

		@Override
		public void startDocument() {
			//
		}

		@Override
		public void startElement(String namespaceURI, String localName, String qName, Attributes atts) {
			if (localName.equals("gruppo")) {
				ultimoGruppo = new LaParolaSegnalibri.GruppoSegnalibri();
				gruppi.add(ultimoGruppo);
				ultimoSegnalibro = null;
				inGruppo = true;
			} else if (localName.equals("sottogruppo")) {
				inSottoGruppo = true;
			} else if (localName.equals("segnalibro")) {
				ultimoSegnalibro = new LaParolaSegnalibri.Segnalibro();
				ultimoGruppo.segnalibri.add(ultimoSegnalibro);
				inSegnalibro = true;
				if (inSottoGruppo && ultimoNomeSottoGruppo != null) {
					ultimoSegnalibro.inizioSottogruppo = ultimoNomeSottoGruppo;
					ultimoNomeSottoGruppo = null;
				}
			} else if (localName.equals("nome")) {
				inNome = true;
			} else if (localName.equals("descrizione")) {
				inDescrizione = true;
			} else if (localName.equals("riferimento")) {
				inRiferimento = true;
			}
		}
	}

	public static void appendLinkGruppo(StringBuilder res, LaParolaSegnalibri.GruppoSegnalibri gruppo) {
		res.append("<p><a href='lpsegnalibro:");
		res.append(gruppo.nome);
		res.append("'>");
		res.append(gruppo.nome);
		res.append("</a>");
		if (gruppo.descrizione != null && gruppo.descrizione.length() > 0) {
			res.append("\n<br />");
			res.append(gruppo.descrizione);
		}
		res.append("</p>\n");
	}

	public static void appendLinkSegnalibro(StringBuilder res, LaParolaSegnalibri.Segnalibro segnalibro, String versionePredefinita) {
		res.append("<p>");
		int n = segnalibro.riferimenti.size();
		if (n == 1 && segnalibro.riferimenti.get(0).contains(":")) {
			res.append("<a href='");
			res.append(segnalibro.riferimenti.get(0));
			res.append("'>");
			res.append(segnalibro.nome);
			res.append("</a>");
		} else {
			res.append(segnalibro.nome);
			res.append(" (");

			for (int i = 0; i < n; i++) {
				String riferimento = segnalibro.riferimenti.get(i);

				String normalizzato = LaParolaBrowser.normalizzaRiferimento(riferimento, versionePredefinita);
				res.append("<a href='laparola:");
				res.append(normalizzato);
				res.append("'>");
				res.append(normalizzato);
				res.append("</a>");

				if (i != n - 1) {
					res.append(", ");
				}
			}

			res.append(")");
		}
		res.append("</p>\n");
	}

	public List<LaParolaSegnalibri.GruppoSegnalibri> gruppi;

	/* package */LaParolaSegnalibri() {
		gruppi = new ArrayList<LaParolaSegnalibri.GruppoSegnalibri>();
	}

	/* package */void aggiungiDaXml(InputStream xml) {
		SegnalibriHandler handler = new SegnalibriHandler();
		SAXParserFactory spf = SAXParserFactory.newInstance();
		SAXParser sp;
		XMLReader xr;
		try {
			sp = spf.newSAXParser();
			xr = sp.getXMLReader();
			xr.setContentHandler(handler);
			xr.parse(new InputSource(xml));
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void aggiungiSegnalibro(String gruppo, String nome, LaParolaUrl url) {
		aggiungiSegnalibro(gruppo, nome, url.getUrl());
	}

	public void aggiungiSegnalibro(String gruppo, String nome, String url) {
		GruppoSegnalibri g = null;

		for (GruppoSegnalibri gg : gruppi) {
			if (gg.nome.equals(gruppo)) {
				g = gg;
				break;
			}
		}

		if (g == null) {
			g = new GruppoSegnalibri(gruppo);
			gruppi.add(g);

			Collections.sort(gruppi);
		}

		g.segnalibri.add(new Segnalibro(nome, url));
		Collections.sort(g.segnalibri);
	}

	/* package */CharSequence salvaInXml() {
		StringBuilder xml = new StringBuilder();

		boolean insottogruppo = false;

		xml.append("<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
		xml.append("<segnalibri>\n");

		for (GruppoSegnalibri g : gruppi) {
			xml.append("  <gruppo>\n");
			xml.append("    <nome>");
			xml.append(g.nome);
			xml.append("</nome>\n");
			xml.append("    <descrizione>");
			xml.append(g.descrizione);
			xml.append("</descrizione>\n");
			for (Segnalibro s : g.segnalibri) {
				if (s.inizioSottogruppo != null) {
					if (insottogruppo) {
						insottogruppo = false;
						xml.append("    </sottogruppo>\n");
					}
					xml.append("    <sottogruppo>\n");
					xml.append("      <nome>");
					xml.append(s.inizioSottogruppo);
					xml.append("</nome>\n");
				}
				xml.append("      <segnalibro>\n");
				xml.append("        <nome>");
				xml.append(s.nome);
				xml.append("</nome>\n");
				for (String r : s.riferimenti) {
					xml.append("        <riferimento>");
					xml.append(r);
					xml.append("</riferimento>\n");
				}
				xml.append("      </segnalibro>\n");
			}

			if (insottogruppo) {
				insottogruppo = false;
				xml.append("    </sottogruppo>\n");
			}
			xml.append("  </gruppo>\n");
		}

		xml.append("</segnalibri>\n");

		return xml;
	}

	public CharSequence getPaginaGruppi(LaParolaBrowser browser) {
		StringBuilder res = new StringBuilder();

		res.append("<html><head>\n");
		if (browser.mClient != null)
			res.append(browser.mClient.getAggiuntaHeader(null));
		res.append("</head><body>\n");

		res.append("<h3>Segnalibri</h3>\n");

		for (LaParolaSegnalibri.GruppoSegnalibri gruppo : gruppi) {
			appendLinkGruppo(res, gruppo);
		}

		res.append("</body></html>\n");

		return res;
	}

	public CharSequence getPaginaGruppo(String gruppo, String versione, LaParolaBrowser browser) {
		StringBuilder res = new StringBuilder();

		res.append("<html><head>\n");
		if (browser.mClient != null)
			res.append(browser.mClient.getAggiuntaHeader(null));
		res.append("</head><body>\n");

		for (LaParolaSegnalibri.GruppoSegnalibri g : gruppi) {
			if (gruppo == null || g.nome.equals(gruppo)) {
				res.append("<h3>");
				res.append(g.nome);
				res.append("</h3>\n");

				if (g.descrizione != null && g.descrizione.length() > 0) {
					res.append("<p>");
					res.append(g.descrizione);
					res.append("</p>\n");
				}

				for (LaParolaSegnalibri.Segnalibro segnalibro : g.segnalibri) {
					if (segnalibro.inizioSottogruppo != null) {
						res.append("<p><b>");
						res.append(segnalibro.inizioSottogruppo);
						res.append("</b></p>\n");
					}
					appendLinkSegnalibro(res, segnalibro, versione);
				}
			}
		}

		res.append("</body></html>\n");

		return res;
	}

	public Segnalibro cercaPerUrl(LaParolaUrl url) {
		// TODO : ha senso fare un Map per la ricerca più rapida?

		if (url == null)
			return null;

		String stringurl = url.getUrlSenzaAncoraggio();

		for (int i = 0; i < gruppi.size(); i++) {
			GruppoSegnalibri g = gruppi.get(i);

			for (int j = 0; j < g.segnalibri.size(); j++) {
				Segnalibro s = g.segnalibri.get(j);

				for (int k = 0; k < s.riferimenti.size(); k++) {
					String r = s.riferimenti.get(k);

					if (r.equals(stringurl)) {
						return s;
					}
				}
			}
		}

		return null;
	}

	public void rimuoviSegnalibro(Segnalibro s) {
		for (int i = 0; i < gruppi.size(); i++) {
			GruppoSegnalibri g = gruppi.get(i);
			g.segnalibri.remove(s);
		}
	}
}